@@ -67,15 +67,16 @@ sub get_lists_for_section {
 
 =head2 get_list
 
-    Returns a specific list by section, name, type, sortscheme
+    Returns a specific list by section, name, type, sortscheme, sortnamekeyscheme
 
 =cut
 
 sub get_list {
-  my ($self, $section, $name, $type, $ssn) = @_;
+  my ($self, $section, $name, $type, $ssn, $snksn) = @_;
   foreach my $list (@{$self->{lists}}) {
     return $list if ($list->get_name eq $name and
                      $list->get_sortschemename eq $ssn and
+                     $list->get_sortnamekeyschemename eq $snksn and
                      $list->get_type eq $type and
                      $list->get_section == $section);
   }
@@ -117,7 +118,7 @@ L<https://github.com/plk/biber/issues>.
 
 =head1 COPYRIGHT & LICENSE
 
-Copyright 2009-2015 François Charette and Philip Kime, all rights reserved.
+Copyright 2009-2016 François Charette and Philip Kime, all rights reserved.
 
 This module is free software.  You can redistribute it and/or
 modify it under the terms of the Artistic License 2.0.
