@@ -4,8 +4,7 @@ use v5.16;
 use strict;
 use Carp;
 use Data::Dump;
-
-use base 'Unicode::Collate::Locale';
+use parent qw(Unicode::Collate::Locale);
 
 my $logger = Log::Log4perl::get_logger('main');
 
@@ -87,7 +86,7 @@ L<https://github.com/plk/biber/issues>.
 
 =head1 COPYRIGHT & LICENSE
 
-Copyright 2009-2015 François Charette and Philip Kime, all rights reserved.
+Copyright 2009-2016 François Charette and Philip Kime, all rights reserved.
 
 This module is free software.  You can redistribute it and/or
 modify it under the terms of the Artistic License 2.0.
